# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Endpoints for uploading and downloading workflow assets"""

from functools import lru_cache
from typing import Union

from fastapi import APIRouter, Header, HTTPException

from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent._shared_files.schemas.asset import AssetSchema
from covalent_dispatcher._object_store.base import TransferDirection

from .._dal.result import get_result_object
from .._db.datastore import workflow_db
from .models import ElectronAssetKey

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

_background_tasks = set()

LRU_CACHE_SIZE = get_config("dispatcher.asset_cache_size")


@router.get("/dispatches/{dispatch_id}/electrons/{node_id}/assets/{key}")
def get_node_asset(
    dispatch_id: str,
    node_id: int,
    key: ElectronAssetKey,
) -> AssetSchema:
    """Returns an asset for an electron.

    Args:
        dispatch_id: The dispatch's unique id.
        node_id: The id of the electron.
        key: The name of the asset
    """

    try:

        result_object = get_cached_result_object(dispatch_id)

        app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")

        node = result_object.lattice.transport_graph.get_node(node_id)
        with workflow_db.session() as session:
            asset = node.get_asset(key=key.value, session=session)
            remote_uri = asset.object_store.get_public_uri(
                asset.storage_path, asset.object_key, direction=TransferDirection.download
            )

        return AssetSchema(size=asset.size, remote_uri=remote_uri)

    except Exception as e:
        app_log.debug(e)
        raise


@router.post("/dispatches/{dispatch_id}/electrons/{node_id}/assets/{key}")
def upload_node_asset(
    dispatch_id: str,
    node_id: int,
    key: ElectronAssetKey,
    content_length: int = Header(default=0),
    digest_alg: Union[str, None] = Header(default=None),
    digest: Union[str, None] = Header(default=None),
) -> AssetSchema:
    """Upload an electron asset.

    Args:
        dispatch_id: The dispatch's unique id.
        node_id: The electron id.
        key: The name of the asset
        asset_file: (body) The file to be uploaded
        content_length: (header)
        digest: (header)
    """
    app_log.debug(
        f"Initiating upload for {dispatch_id}:{node_id}:{key.value} ({content_length} bytes) "
    )
    try:
        metadata = {"size": content_length, "digest_alg": digest_alg, "digest": digest}
        remote_uri = _update_node_asset_metadata(
            dispatch_id,
            node_id,
            key,
            metadata,
        )
        return AssetSchema(size=content_length, remote_uri=remote_uri)
    except Exception as e:
        app_log.debug(e)
        raise


# This must only be used for static data as we don't have yet any
# intelligent invalidation logic.
@lru_cache(maxsize=LRU_CACHE_SIZE)
def get_cached_result_object(dispatch_id: str):
    try:
        with workflow_db.session() as session:
            srv_res = get_result_object(dispatch_id, bare=False, session=session)
            app_log.debug(f"Caching result {dispatch_id}")

            # Prepopulate asset maps to avoid DB lookups

            srv_res.populate_asset_map(session)
            srv_res.lattice.populate_asset_map(session)

            tg = srv_res.lattice.transport_graph
            g = tg.get_internal_graph_copy()
            for node_id in g.nodes():
                node = tg.get_node(node_id, session)
                node.populate_asset_map(session)
    except KeyError:
        raise HTTPException(
            status_code=404,
            detail=f"The requested dispatch ID {dispatch_id} was not found.",
        )

    return srv_res


def _filter_null_metadata(metadata):
    # Filter out null updates
    return {k: v for k, v in metadata.items() if v is not None}


def _update_node_asset_metadata(dispatch_id, node_id, key, metadata) -> str:
    result_object = get_cached_result_object(dispatch_id)

    app_log.debug(f"LRU cache info: {get_cached_result_object.cache_info()}")
    node = result_object.lattice.transport_graph.get_node(node_id)
    with workflow_db.session() as session:
        asset = node.get_asset(key=key.value, session=session)
        app_log.debug(f"Asset uri {asset.internal_uri}")

        # Update asset metadata
        update = _filter_null_metadata(metadata)
        node.update_assets(updates={key: update}, session=session)
        app_log.debug(f"Updated node asset {dispatch_id}:{node_id}:{key.value}")

        object_store = asset.object_store
        remote_uri = object_store.get_public_uri(
            asset.storage_path, asset.object_key, direction=TransferDirection.upload
        )
        return remote_uri



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from pathlib import Path
from typing import Optional

from furl import furl

from .enums import FileSchemes

_is_remote_scheme = {
    FileSchemes.S3.value: True,
    FileSchemes.Blob.value: True,
    FileSchemes.GCloud.value: True,
    FileSchemes.Globus.value: True,
    FileSchemes.HTTP.value: True,
    FileSchemes.HTTPS.value: True,
    FileSchemes.FTP.value: True,
    FileSchemes.File: False,
}


# For registering additional file transfer strategies; this will be called by
# `register_uploader`` and `register_downloader``
def register_remote_scheme(s: str):
    _is_remote_scheme[s] = True


class File:
    """
    File class to store components of provided URI including scheme (s3://, file://, ect.) determine if the file is remote,
    and acts a facade to facilitate filesystem operations.

    Attributes:
        filepath: File path corresponding to the file.
        is_remote: Flag determining if file is remote (override). Default is resolved automatically from file scheme.
        is_dir: Flag determining if file is a directory (override). Default is determined if file uri contains trailing slash.
        include_folder: Flag that determines if the folder should be included in the file transfer, if False only contents of folder are transfered.
    """

    _is_remote = False
    _is_dir = False
    _include_folder = False

    def __init__(
        self,
        filepath: Optional[str] = None,
        is_remote: bool = False,
        is_dir: bool = False,
        include_folder: bool = False,
    ):
        if not isinstance(filepath, str) and filepath is not None:
            raise AttributeError(
                "Only strings are valid filepaths for a covalent File constructor."
            )

        self.id = str(uuid.uuid4())

        # assign default filepath of form /tmp/{id}
        if filepath is None:
            filepath = self.get_temp_filepath()

        # override is_remote boolean
        if is_remote:
            self._is_remote = True

        if is_dir:
            self._is_dir = True

        if include_folder:
            self._include_folder = True

        self.scheme = File.resolve_scheme(filepath)
        self._path_object = File.get_path_obj(filepath)
        self.uri = File.get_uri(self.scheme, filepath)

    @property
    def is_temp_file(self):
        return self.filepath == self.get_temp_filepath()

    def get_temp_filepath(self):
        return f"/tmp/{self.id}"

    @property
    def is_remote(self):
        return self._is_remote or _is_remote_scheme[self.scheme]

    @property
    def filepath(self) -> str:
        path_obj = self._path_object
        if not self._include_folder and self.is_dir:
            path_obj = path_obj / "/"
        path_obj.normalize()
        filepath = str(path_obj)
        if self._include_folder:
            filepath = filepath.rstrip("/")
        return filepath

    @property
    def is_dir(self):
        return self._is_dir or self._path_object.isdir

    def touch(self):
        Path(self.filepath).touch()

    @staticmethod
    def get_path_obj(path: str):
        path_components = furl(path)
        path_components.scheme = None
        path_components.path.normalize()
        return path_components.path

    @staticmethod
    def get_uri(scheme: str, path: str) -> str:
        path_components = furl(path)
        path_components.scheme = scheme
        path_components.path.normalize()
        return path_components.url

    @staticmethod
    def resolve_scheme(path: str) -> str:
        scheme = furl(path).scheme
        host = furl(path).host
        # Canonicalize file system paths to file:// urls
        if not scheme:
            return FileSchemes.File.value
        if scheme in _is_remote_scheme:
            return scheme
        else:
            raise ValueError(f"Provided File scheme ({scheme}) is not supported.")



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import requests

from ..._shared_files import logger
from .. import File
from .transfer_strategy_base import FileTransferStrategy

app_log = logger.app_log


class HTTP(FileTransferStrategy):
    """
    Implements Base FileTransferStrategy class to use download files from public http(s) URLs.
    """

    # return callable to download here implies 'from' is a remote source
    def download(self, from_file: File, to_file: File = File()) -> File:
        from_filepath = from_file.uri
        to_filepath = to_file.filepath

        def callable():
            resp = requests.get(from_filepath, stream=True)
            resp.raise_for_status()
            with open(to_filepath, "wb") as f:
                for chunk in resp.iter_content(chunk_size=1024):
                    f.write(chunk)

        return callable

    # Upload a file to a (possibly presigned) HTTP(s) URL
    def upload(self, from_file: File, to_file: File = File()) -> File:
        from_filepath = from_file.filepath
        to_filepath = to_file.uri
        filesize = os.path.getsize(from_filepath)

        def callable():
            with open(from_filepath, "rb") as reader:
                # Workaround for Requests bug when streaming from empty files
                app_log.debug(f"uploading to {to_filepath}")
                data = reader.read() if filesize < 50 else reader
                r = requests.put(to_filepath, headers={"Content-Length": str(filesize)}, data=data)
                r.raise_for_status()

        return callable

    # HTTP Strategy does not support server-side copy between two remote URLs
    def cp(self, from_file: File, to_file: File = File()) -> File:
        raise NotImplementedError



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union

from .enums import FileSchemes, FtCallDepReturnValue, Order
from .file import File, register_remote_scheme
from .strategies.blob_strategy import Blob
from .strategies.gcloud_strategy import GCloud
from .strategies.http_strategy import HTTP
from .strategies.s3_strategy import S3
from .strategies.shutil_strategy import Shutil
from .strategies.transfer_strategy_base import FileTransferStrategy

_downloaders = {
    FileSchemes.File.value: Shutil,
    FileSchemes.HTTP.value: HTTP,
    FileSchemes.HTTPS.value: HTTP,
    FileSchemes.GCloud.value: GCloud,
    FileSchemes.Blob.value: Blob,
    FileSchemes.S3.value: S3,
}

_uploaders = {
    FileSchemes.File.value: Shutil,
    FileSchemes.HTTP.value: HTTP,
    FileSchemes.HTTPS.value: HTTP,
    FileSchemes.GCloud.value: GCloud,
    FileSchemes.Blob.value: Blob,
    FileSchemes.S3.value: S3,
}


# For registering additional file transfer strategies
def register_downloader(scheme: str, cls: FileTransferStrategy):
    register_remote_scheme(scheme)
    _downloaders[scheme] = cls


def register_uploader(scheme: str, cls: FileTransferStrategy):
    register_remote_scheme(scheme)
    _uploaders[scheme] = cls


def guess_transfer_strategy(from_file: File, to_file: File) -> FileTransferStrategy:
    # Handle the following cases automatically
    # Local-Remote (except HTTP destination)
    # Remote-local
    # Local-local

    # Local-Remote
    if not from_file.is_remote and to_file.is_remote:
        strategy = _uploaders.get(to_file.scheme)
        if not strategy:
            raise AttributeError(f"Cannot guess upload strategy for remote {to_file.uri}")
        return strategy

    # Remote-Local
    if from_file.is_remote and not to_file.is_remote:
        strategy = _downloaders.get(from_file.scheme)
        if not strategy:
            raise AttributeError(f"Cannot guess download strategy for remote {from_file.uri}")
        return strategy

    # Local-Local
    if not from_file.is_remote and not to_file.is_remote:
        strategy = Shutil
        return strategy

    else:
        raise AttributeError("FileTransfer requires a file transfer strategy to be specified")


class FileTransfer:
    """
    FileTransfer object class that takes two File objects or filepaths (from, to) and a File Transfer Strategy to perform remote or local file transfer operations.

    Attributes:
        from_file: Filepath or File object corresponding to the source file.
        to_file: Filepath or File object corresponding to the destination file.
        order: Order (enum) to execute the file transfer before (Order.BEFORE) or after (Order.AFTER) electron execution.
        strategy: Optional File Transfer Strategy to perform file operations - default will be resolved from provided file schemes.
    """

    def __init__(
        self,
        from_file: Optional[Union[File, str]] = None,
        to_file: Optional[Union[File, str]] = None,
        order: Order = Order.BEFORE,
        strategy: Optional[FileTransferStrategy] = None,
    ) -> None:
        if isinstance(from_file, str) or from_file is None:
            from_file = File(from_file)
        elif not isinstance(from_file, File):
            raise AttributeError(
                "FileTransfer() requires files to be either of type File, string, or None."
            )

        if isinstance(to_file, str) or to_file is None:
            to_file = File(to_file)
        elif not isinstance(to_file, File):
            raise AttributeError(
                "FileTransfer requires files to be either of type File, string, or None."
            )

        # assign explicit strategy or default to strategy based on from_file & to_file schemes
        if strategy:
            self.strategy = strategy
        else:
            self.strategy = guess_transfer_strategy(from_file, to_file)()

        self.to_file = to_file
        self.from_file = from_file
        self.order = order

        # this is currently the case but as we further develop file transfer strategies we may support this
        # for example we may support streaming files between buckets in S3
        if self.from_file.is_remote and self.to_file.is_remote:
            raise ValueError(
                "FileTransfer currently does not support remote->remote file transfers, please update from_filepath or to_filepath to correspond to a local filepath."
            )

    def cp(self):
        file_transfer_call_dep = None
        return_value_type = FtCallDepReturnValue.FROM_TO

        # local -> local or remote -> remote
        if (not self.from_file.is_remote and not self.to_file.is_remote) or (
            self.from_file.is_remote and self.to_file.is_remote
        ):
            file_transfer_call_dep = self.strategy.cp(self.from_file, self.to_file)
        # local -> remote
        if not self.from_file.is_remote and self.to_file.is_remote:
            file_transfer_call_dep = self.strategy.upload(self.from_file, self.to_file)
        # remote -> local
        if self.from_file.is_remote and not self.to_file.is_remote:
            file_transfer_call_dep = self.strategy.download(self.from_file, self.to_file)

        pre_transfer_hook_call_dep = self.strategy.pre_transfer_hook(
            self.from_file, self.to_file, return_value_type=return_value_type
        )

        return (pre_transfer_hook_call_dep, file_transfer_call_dep)


# Factories


def TransferFromRemote(
    from_filepath: str,
    to_filepath: Union[str, None] = None,
    strategy: Optional[FileTransferStrategy] = None,
    order: Order = Order.BEFORE,
) -> FileTransfer:
    """
    Factory for creating a FileTransfer instance where from_filepath is implicitly created as a remote File Object, and the order (Order.BEFORE) is set so that this file transfer will occur prior to electron execution.

    Args:
        from_filepath: File path corresponding to remote file (source).
        to_filepath: File path corresponding to local file (destination)
        strategy: Optional File Transfer Strategy to perform file operations - default will be resolved from provided file schemes.
        order: Order (enum) to execute the file transfer before (Order.BEFORE) or after (Order.AFTER) electron execution - default is BEFORE

    Returns:
        FileTransfer instance with implicit Order.BEFORE enum set and from (source) file marked as remote
    """
    # override is_remote for the case where from_filepath is of a file:// scheme where the file is remote (rsync ssh)
    is_dir = from_filepath.endswith("/")
    from_file = File(from_filepath, is_remote=True, is_dir=is_dir)
    if is_dir and not to_filepath.endswith("/"):
        raise ValueError("Please specify a directory path (ending with '/') as the destination.")
    to_file = File(to_filepath, is_dir=is_dir)
    return FileTransfer(
        from_file=from_file, to_file=to_file, order=Order.BEFORE, strategy=strategy
    )


def TransferToRemote(
    to_filepath: str,
    from_filepath: Union[str, None] = None,
    strategy: Optional[FileTransferStrategy] = None,
    order: Order = Order.AFTER,
) -> FileTransfer:
    """
    Factory for creating a FileTransfer instance where to_filepath is implicitly created as a remote File Object, and the order (Order.AFTER) is set so that this file transfer will occur post electron execution.

    Args:
        to_filepath: File path corresponding to remote file (destination)
        from_filepath: File path corresponding to local file (source).
        strategy: Optional File Transfer Strategy to perform file operations - default will be resolved from provided file schemes.
        order: Order (enum) to execute the file transfer before (Order.BEFORE) or after (Order.AFTER) electron execution - default is AFTER

    Returns:
        FileTransfer instance with implicit Order.AFTER enum set and to (destination) file marked as remote
    """
    # override is_remote for the case where to_filepath is of a file:// scheme where the file is remote (rsync ssh)
    is_dir = from_filepath and from_filepath.endswith("/")
    from_file = File(from_filepath, is_dir=is_dir)
    if is_dir and not to_filepath.endswith("/"):
        raise ValueError("Please specify a directory path (ending with '/') as the destination.")
    to_file = File(to_filepath, is_remote=True, is_dir=is_dir)

    return FileTransfer(from_file=from_file, to_file=to_file, order=Order.AFTER, strategy=strategy)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add task_group table

Revision ID: 7e9fb153ecfb
Revises: 7b40a55c0752
Create Date: 2025-03-25 16:36:49.408467

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
# pragma: allowlist nextline secret
revision = "7e9fb153ecfb"
# pragma: allowlist nextline secret
down_revision = "7b40a55c0752"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "taskgroupstate",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("dispatch_id", sa.Text(), nullable=False),
        sa.Column("task_group_id", sa.Integer(), nullable=False),
        sa.Column("num_pending_parents", sa.Integer(), nullable=False),
        sa.Column("sorted_tasks", sa.Text(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "workflowstate",
        sa.Column("dispatch_id", sa.Text(), nullable=False),
        sa.Column("num_unresolved_tasks", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("dispatch_id"),
    )

    with op.batch_alter_table("taskgroupstate", schema=None) as batch_op:
        batch_op.create_index("task_grp_idx", ["dispatch_id", "task_group_id"], unique=False)

    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.create_unique_constraint("u_electron_id", ["electron_id"])

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("lattices", schema=None) as batch_op:
        batch_op.drop_constraint("u_electron_id", type_="unique")

    with op.batch_alter_table("taskgroupstate", schema=None) as batch_op:
        batch_op.drop_index("task_grp_idx")

    op.drop_table("workflowstate")
    op.drop_table("taskgroupstate")
    # ### end Alembic commands ###



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Endpoints for dispatch management"""

import asyncio
import json
from contextlib import asynccontextmanager
from typing import Annotated, List, Union

from fastapi import APIRouter, FastAPI, HTTPException, Query
from fastapi.responses import JSONResponse

import covalent_dispatcher.entry_point as dispatcher
from covalent._shared_files import logger
from covalent._shared_files.schemas.result import ResultSchema
from covalent._shared_files.util_classes import RESULT_STATUS
from covalent_dispatcher._core import dispatcher as core_dispatcher
from covalent_dispatcher._core import runner_ng as core_runner

from .._dal.exporters.result import export_result_manifest
from .._dal.result import Result, get_result_object
from .._db.datastore import workflow_db
from .._db.dispatchdb import DispatchDB
from .heartbeat import Heartbeat
from .models import (
    BulkDispatchGetSchema,
    BulkGetMetadata,
    DispatchStatusSetSchema,
    DispatchSummary,
    ElectronUpdateSchema,
    TargetDispatchStatus,
)

app_log = logger.app_log
log_stack_info = logger.log_stack_info

router: APIRouter = APIRouter()

_background_tasks = set()


@asynccontextmanager
async def lifespan(app: FastAPI):
    """Initialize global variables"""

    heartbeat = Heartbeat()
    fut = asyncio.create_task(heartbeat.start())
    _background_tasks.add(fut)
    fut.add_done_callback(_background_tasks.discard)

    # Runner event queue and listener
    core_runner._job_events = asyncio.Queue()
    core_runner._job_event_listener = asyncio.create_task(core_runner._listen_for_job_events())

    # Dispatcher event queue and listener
    core_dispatcher._global_status_queue = asyncio.Queue()
    core_dispatcher._global_event_listener = asyncio.create_task(
        core_dispatcher._node_event_listener()
    )

    yield

    # Cancel all scheduled and running dispatches
    for status in [
        RESULT_STATUS.NEW_OBJECT,
        RESULT_STATUS.RUNNING,
    ]:
        await cancel_all_with_status(status)

    core_dispatcher._global_event_listener.cancel()
    core_runner._job_event_listener.cancel()

    Heartbeat.stop()


async def cancel_all_with_status(status: RESULT_STATUS):
    """Cancel all dispatches with the specified status."""

    with workflow_db.session() as session:
        records = Result.get_db_records(
            session,
            keys=["dispatch_id"],
            equality_filters={"status": str(status)},
            membership_filters={},
        )

        for record in records:
            dispatch_id = record.dispatch_id
            await dispatcher.cancel_running_dispatch(dispatch_id)


async def start(dispatch_id: str):
    """Start a previously registered (re-)dispatch.

    Args:
        `dispatch_id`: The dispatch's unique id.

    Returns:
        `dispatch_id`
    """
    fut = asyncio.create_task(dispatcher.start_dispatch(dispatch_id))
    _background_tasks.add(fut)
    fut.add_done_callback(_background_tasks.discard)

    return dispatch_id


async def cancel(dispatch_id: str, task_ids: List[int] = None) -> str:
    """
    Function to handle the cancel request of
    a dispatch.

    Args:
        dispatch_id: ID of the dispatch
        task_ids: (Query) Optional list of specific task ids to cancel.
            An empty list will cause all tasks to be cancelled.

    Returns:
        Fast API Response object confirming that the dispatch
        has been cancelled.
    """

    if task_ids is None:
        task_ids = []

    await dispatcher.cancel_running_dispatch(dispatch_id, task_ids)
    if task_ids:
        return f"Cancelled tasks {task_ids} in dispatch {dispatch_id}."
    else:
        return f"Dispatch {dispatch_id} cancelled."


@router.get("/db-path")
def db_path() -> str:
    db_path = DispatchDB()._dbpath
    return json.dumps(db_path)


@router.post("/dispatches", status_code=201)
async def register(manifest: ResultSchema) -> ResultSchema:
    """Register a dispatch in the database.

    Args:
        manifest: Declares all metadata and assets in the workflow
        parent_dispatch_id: The parent dispatch id if registering a sublattice dispatch

    Returns:
        The manifest with `dispatch_id` and remote URIs for each asset populated.
    """
    try:
        return await dispatcher.register_dispatch(manifest, None)
    except Exception as e:
        app_log.debug(f"Exception in register: {e}")
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


@router.patch("/dispatches/{dispatch_id}/electrons/{node_id}")
def associate_sublattice_with_electron(dispatch_id: str, node_id: str, body: ElectronUpdateSchema):
    dispatch_controller = Result.meta_type

    with workflow_db.session() as session:
        res = get_result_object(dispatch_id, bare=True, session=session)
        node = res.lattice.transport_graph.get_node(node_id, session=session)
        electron_id = node._electron_id

        dispatch_controller.update_bulk(
            session=session,
            values={"electron_id": electron_id},
            equality_filters={"dispatch_id": body.sub_dispatch_id},
            membership_filters={},
        )
        session.commit()

    app_log.debug(
        f"Associated sublattice {body.sub_dispatch_id} with parent electron {dispatch_id}:{node_id}"
    )


@router.post("/dispatches/{dispatch_id}/redispatches", status_code=201)
async def register_redispatch(
    manifest: ResultSchema,
    dispatch_id: str,
    reuse_previous_results: bool = False,
):
    """Register a redispatch in the database.

    Args:
        manifest: Declares all metadata and assets in the workflow
        dispatch_id: The original dispatch's id.
        reuse_previous_results: Whether to try reusing the results of
            previously completed electrons.

    Returns:
        The manifest with `dispatch_id` and remote URIs for each asset populated.
    """
    try:
        return await dispatcher.register_redispatch(
            manifest,
            dispatch_id,
            reuse_previous_results,
        )
    except Exception as e:
        app_log.debug(f"Exception in register_redispatch: {e}")
        raise HTTPException(
            status_code=400,
            detail=f"Failed to submit workflow: {e}",
        ) from e


@router.put("/dispatches/{dispatch_id}/status", status_code=202)
async def set_dispatch_status(dispatch_id: str, desired_status: DispatchStatusSetSchema):
    """Set the status of a dispatch.

    Valid target statuses are:
        - "RUNNING" to start a dispatch
        - "CANCELLED" to cancel dispatch processing

    Args:
        `dispatch_id`: The dispatch's unique id
        `desired_status`: A `StatusSetSchema` object describing the desired status.

    """

    if desired_status.status == TargetDispatchStatus.running:
        return await start(dispatch_id)
    else:
        return await cancel(dispatch_id, desired_status.task_ids)


@router.get("/dispatches", response_model_exclude_unset=True)
def get_dispatches_bulk(
    dispatch_id: Annotated[Union[List[str], None], Query()] = None,
    status: Annotated[Union[List[str], None], Query()] = None,
    latest: bool = True,
    offset: int = 0,
    count: int = 10,
    status_only: bool = False,
) -> BulkDispatchGetSchema:
    dispatch_controller = Result.meta_type

    if status_only:
        fields = ["dispatch_id", "status"]
    else:
        fields = [
            "dispatch_id",
            "root_dispatch_id",
            "status",
            "name",
            "electron_num",
            "completed_electron_num",
            "created_at",
            "updated_at",
            "completed_at",
        ]

    summaries = []
    with workflow_db.session() as session:
        in_filters = {}
        if dispatch_id is not None:
            in_filters["dispatch_id"] = dispatch_id
        if status is not None:
            in_filters["status"] = status

        results = dispatch_controller.get(
            session,
            fields=fields,
            equality_filters={"is_active": True},
            membership_filters=in_filters,
            sort_fields=["created_at"],
            reverse=latest,
            offset=offset,
            max_items=count,
        )
        for res in results:
            dispatch_id = res.dispatch_id
            summary = DispatchSummary.model_validate(res)
            summaries.append(summary)

    bulk_meta = BulkGetMetadata(count=len(results))
    return BulkDispatchGetSchema(dispatches=summaries, metadata=bulk_meta)


@router.get("/dispatches/{dispatch_id}")
def export_manifest(dispatch_id: str) -> ResultSchema:
    result_object = _try_get_result_object(dispatch_id)
    if not result_object:
        return JSONResponse(
            status_code=404,
            content={"message": f"The requested dispatch ID {dispatch_id} was not found."},
        )

    return export_result_manifest(dispatch_id)


def _try_get_result_object(dispatch_id: str) -> Union[Result, None]:
    try:
        res = get_result_object(dispatch_id, bare=True)
    except KeyError:
        res = None
    return res



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper classes for the dispatcher
"""

import json
import os
import tempfile

from covalent_dispatcher._core.data_modules.utils import run_in_executor
from covalent_dispatcher._dal.base import workflow_db
from covalent_dispatcher._dal.dispatcher_state import TaskGroupState, WorkflowState
from covalent_dispatcher._db.datastore import DataStore


class _WorkflowRunState:

    def __init__(self, db: DataStore):
        self.db = db

    def _get_unresolved(self, dispatch_id: str):
        with self.db.session() as session:
            records = WorkflowState.get(
                session,
                fields=["num_unresolved_tasks"],
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            return records[0].num_unresolved_tasks

    async def get_unresolved(self, dispatch_id: str):
        return await run_in_executor(self._get_unresolved, dispatch_id)

    def _set_unresolved(self, dispatch_id: str, val: int):
        with self.db.session() as session:
            WorkflowState.create(
                session,
                insert_kwargs={
                    "dispatch_id": dispatch_id,
                    "num_unresolved_tasks": val,
                },
            )
            session.commit()

    async def set_unresolved(self, dispatch_id: str, val: int):
        return await run_in_executor(self._set_unresolved, dispatch_id, val)

    def _increment(self, dispatch_id: str, interval: int = 1):
        with self.db.session() as session:
            WorkflowState.incr_bulk(
                session=session,
                increments={"num_unresolved_tasks": interval},
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            records = WorkflowState.get(
                session,
                fields=["num_unresolved_tasks"],
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            session.commit()
            return records[0].num_unresolved_tasks

    async def increment(self, dispatch_id: str, interval: int = 1):
        return await run_in_executor(self._increment, dispatch_id, interval)

    def _decrement(self, dispatch_id: str):
        with self.db.session() as session:
            WorkflowState.incr_bulk(
                session=session,
                increments={"num_unresolved_tasks": -1},
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            records = WorkflowState.get(
                session,
                fields=["num_unresolved_tasks"],
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            session.commit()
            return records[0].num_unresolved_tasks

    async def decrement(self, dispatch_id: str):
        return await run_in_executor(self._decrement, dispatch_id)

    def _remove(self, dispatch_id: str):
        with self.db.session() as session:
            WorkflowState.delete_bulk(
                session=session,
                equality_filters={"dispatch_id": dispatch_id},
                membership_filters={},
            )
            session.commit()

    async def remove(self, dispatch_id: str):
        await run_in_executor(self._remove, dispatch_id)


class TaskGroupRunState:

    def __init__(self, db):
        self.db = db

    def _get_pending(self, dispatch_id: str, task_group_id: int):
        with self.db.session() as session:
            records = TaskGroupState.get(
                session=session,
                fields=["num_pending_parents"],
                equality_filters={"dispatch_id": dispatch_id, "task_group_id": task_group_id},
                membership_filters={},
            )
            return records[0].num_pending_parents

    async def get_pending(self, dispatch_id: str, task_group_id: int):
        return await run_in_executor(self._get_pending, dispatch_id, task_group_id)

    def _set(self, dispatch_id: str, task_group_id: int, num_pending: int, sorted_nodes):
        with self.db.session() as session:
            TaskGroupState.create(
                session=session,
                insert_kwargs={
                    "dispatch_id": dispatch_id,
                    "task_group_id": task_group_id,
                    "num_pending_parents": num_pending,
                    "sorted_tasks": json.dumps(sorted_nodes),
                },
            )
            session.commit()

    async def set(self, dispatch_id: str, task_group_id: int, num_pending: int, sorted_nodes):
        return await run_in_executor(
            self._set, dispatch_id, task_group_id, num_pending, sorted_nodes
        )

    def _decrement(self, dispatch_id: str, task_group_id):
        with self.db.session() as session:
            TaskGroupState.incr_bulk(
                session=session,
                increments={"num_pending_parents": -1},
                equality_filters={"dispatch_id": dispatch_id, "task_group_id": task_group_id},
                membership_filters={},
            )
            records = TaskGroupState.get(
                session,
                fields=["num_pending_parents"],
                equality_filters={"dispatch_id": dispatch_id, "task_group_id": task_group_id},
                membership_filters={},
            )
            session.commit()
            return records[0].num_pending_parents

    async def decrement(self, dispatch_id: str, task_group_id: int):
        return await run_in_executor(self._decrement, dispatch_id, task_group_id)

    async def remove(self, dispatch_id: str, task_group_id: int):
        pass

    def _get_task_group(self, dispatch_id: str, task_group_id: int):
        with self.db.session() as session:
            records = TaskGroupState.get(
                session=session,
                fields=["sorted_tasks"],
                equality_filters={"dispatch_id": dispatch_id, "task_group_id": task_group_id},
                membership_filters={},
            )
            return json.loads(records[0].sorted_tasks)

    async def get_task_group(self, dispatch_id: str, task_group_id: int):
        return await run_in_executor(self._get_task_group, dispatch_id, task_group_id)

    def _remove(self, dispatch_id: str, task_group_id: int):
        with self.db.session() as session:
            TaskGroupState.delete_bulk(
                session=session,
                equality_filters={"dispatch_id": dispatch_id, "task_group_id": task_group_id},
                membership_filters={},
            )
            session.commit()

    async def remove(self, dispatch_id: str, task_group_id: int):
        await run_in_executor(self._remove, dispatch_id, task_group_id)


# Default to tmpfs backed file
cache_db_file = tempfile.NamedTemporaryFile(
    mode="w+b", prefix="covalent-dispatcher-cache-", suffix=".db"
)
cache_db_URL = os.environ.get("COVALENT_CACHE_DB_URL", f"sqlite+pysqlite:///{cache_db_file.name}")
initialize_db = True

# If we want to store dispatcher state in the main DB, let the alembic migrations
# create the tables
if cache_db_URL == workflow_db.db_URL:
    initialize_db = False

cache_db = DataStore(db_URL=cache_db_URL, initialize_db=initialize_db)

_task_group_cache = TaskGroupRunState(db=cache_db)
_workflow_run_cache = _WorkflowRunState(db=cache_db)



# Copyright 2024 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Embedded file server API"""

import os

import aiofiles
import aiofiles.os
from fastapi import APIRouter, Request
from fastapi.exceptions import RequestValidationError
from fastapi.responses import FileResponse

from covalent._shared_files import logger
from covalent._shared_files.config import get_config

router = APIRouter()

app_log = logger.app_log
BASE_PATH = get_config("dispatcher.results_dir")


async def _transfer_data(req: Request, dest_path: str):

    # Stream data to a temporary file, then replace the destination
    # file atomically
    tmp_path = f"{dest_path}.tmp"
    app_log.debug(f"Streaming file upload to {tmp_path}")

    async with aiofiles.open(tmp_path, "wb") as f:
        async for chunk in req.stream():
            await f.write(chunk)

    await aiofiles.os.replace(tmp_path, dest_path)


# Resolve path to an absolute path and check that it
# doesn't escape the data directory root
def _sanitize_path(path: str) -> str:
    abs_path = os.path.realpath(path)
    if not abs_path.startswith(BASE_PATH) or len(abs_path) <= len(BASE_PATH):
        raise RequestValidationError(f"Invalid object key {path}")
    return abs_path


@router.get("/files/{object_key:path}")
async def download_file(object_key: str):
    path = _sanitize_path(os.path.join(BASE_PATH, object_key))
    return FileResponse(path)


@router.put("/files/{object_key:path}")
async def upload_file(req: Request, object_key: str):
    path = _sanitize_path(os.path.join(BASE_PATH, object_key))
    os.makedirs(os.path.dirname(path), exist_ok=True)
    await _transfer_data(req, path)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import shutil
from dataclasses import asdict
from functools import reduce
from operator import getitem
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import filelock
import toml

"""Configuration manager."""


class ConfigManager:
    """
    Configuration manager class object.

    This class is used to manage an in-memory copy of a TOML configuration file.
    """

    def __init__(self) -> None:
        from .defaults import DefaultConfig

        DEFAULT_CONFIG = asdict(DefaultConfig())

        self.config_file = DEFAULT_CONFIG["sdk"]["config_file"]

        self.generate_default_config()

        try:
            with open(self.config_file, "r"):
                pass

            self.update_config()
        except FileNotFoundError:
            Path(self.config_file).parent.mkdir(parents=True, exist_ok=True)
            self.write_config()

        Path(self.get("sdk.log_dir")).mkdir(parents=True, exist_ok=True)
        Path(self.get("sdk.executor_dir")).mkdir(parents=True, exist_ok=True)
        Path(self.get("dispatcher.db_path")).parent.mkdir(parents=True, exist_ok=True)

    def generate_default_config(self) -> None:
        """
        Load the default configuration into memory.

        Args:
            None

        Returns:
            None
        """

        from .defaults import DefaultConfig

        self.config_data = asdict(DefaultConfig())

    def update_config(
        self, new_entries: Optional[Dict] = None, override_existing: bool = True
    ) -> None:
        """
        Update the exising configuration dictionary with the configuration stored in file.
            Optionally, update configuration data with an input dict.

        Args:
            new_entries: Dictionary of new entries added or updated in the config.
            override_existing: If True (default), config values from the config file
                or the input dictionary (new_entries) take precedence over any existing
                values in the config.

        Returns:
            None
        """

        def update_nested_dict(old_dict, new_dict, override_existing: bool = True):
            for key, value in new_dict.items():
                if isinstance(value, dict) and key in old_dict and isinstance(old_dict[key], dict):
                    update_nested_dict(old_dict[key], value, override_existing)
                else:
                    if override_existing:
                        # Values provided should override existing values.
                        old_dict[key] = value
                    else:
                        # Values provided are defaults, and shouldn't override existing values
                        # unless the existing value is empty.
                        if key in old_dict and old_dict[key] == "":
                            old_dict[key] = value
                        else:
                            old_dict.setdefault(key, value)

        with filelock.FileLock(f"{self.config_file}.lock", timeout=1):
            with open(self.config_file, "r+") as f:
                file_config = toml.load(f)

                update_nested_dict(self.config_data, file_config)
                if new_entries:
                    update_nested_dict(self.config_data, new_entries, override_existing)

                # Writing it back to the file
                self.write_config()

    def read_config(self) -> None:
        """
        Read the configuration from file.

        Args:
            None

        Returns:
            None
        """

        self.config_data = toml.load(self.config_file)

    def write_config(self) -> None:
        """
        Write the configuration to file.

        Args:
            None

        Returns:
            None
        """

        with open(self.config_file, "w") as f:
            toml.dump(self.config_data, f)

    def purge_config(self) -> None:
        """
        Purge the configuration directory.

        Args:
            None

        Returns:
            None
        """

        dir_name = os.path.dirname(self.config_file)
        shutil.rmtree(dir_name, ignore_errors=True)

    def get(self, key: str) -> Any:
        """
        Return a value given a dictionary key.

        Args:
            key: Key value in period-delimited format, e.g., config[dispatcher][port]
                 is queried by passing "dispatcher.port" as the key.

        Returns:
            value: The corresponding configuration value, usually a string or int.
        """

        return reduce(getitem, key.split("."), self.config_data)

    def set(self, key: str, value: Any) -> None:
        """
        Set a key-value pair in the configuration dictionary.

        Args:
            key: Key value in period-delimited format.
            value: The corresponding configuration setting.

        Returns:
            None
        """

        keys = key.split(".")

        try:
            reduce(getitem, keys[:-1], self.config_data)[keys[-1]] = value
        except KeyError:
            data = self.config_data
            for key in keys:
                data[key] = {}
                data = data[key]
            data[keys[-1]] = value


_config_manager = ConfigManager()


def set_config(new_config: Union[Dict, str], new_value: Any = None) -> None:
    """
    Update the configuration.

    Users may pass a dictionary of new settings, or a string key with a value to set
    a single configuration setting.

    Args:
        new_config: The new configuration dictionary, or a string key name.
        new_value: A new configuration value, if the first argument is a string.

    Returns:
        None
    """
    cm = _config_manager

    if isinstance(new_config, str):
        cm.set(new_config, new_value)
    else:
        for key, value in new_config.items():
            cm.set(key, value)
    cm.write_config()


def get_config(entries: Union[str, List] = None) -> Union[Dict, Union[str, int]]:
    """
    Return a configuration setting.

    Invocation with no arguments returns the full configuration description; with a
    list of arguments returns a dictionary of configuration settings; with a string
    key name returns the corresponding value for a single setting.

    Args:
        entries: A string or list of strings specifying key names.

    Returns:
        config: A dictionary or string describing the corresponding configuration
                settings.
    """

    entries = entries or []
    cm = _config_manager

    if isinstance(entries, List) and len(entries) == 0:
        # If no arguments are passed, return the full configuration as a dict
        return cm.config_data
    elif isinstance(entries, List) and len(entries) == 1:
        # If the argument is a single key in a List, return the corresponding value
        return cm.get(entries[0])
    elif isinstance(entries, str):
        # If the argument is a string key, return the corresponding value
        return cm.get(entries)
    else:
        # If a set of keys are passed, return a corresponding dict of key-value pairs
        values = [cm.get(entry) for entry in entries]
        return dict(zip(entries, values))


def reload_config() -> None:
    """
    Reload the configuration from the TOML file.

    Args:
        None

    Returns:
        None
    """
    cm = _config_manager
    cm.read_config()


def update_config(new_entries: Optional[Dict] = None, override_existing: bool = True) -> None:
    """
    Read the configuration from the TOML file and append to default
        (or existing) configuration. Optionally, update configuration
        data with an input dict.

    Args:
        new_entries: Dictionary of new entries added or updated in the config
        override_existing: If True (which is the default), default values overwrite
            existing entries.

    Returns:
        None
    """
    cm = _config_manager
    cm.update_config(new_entries, override_existing)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for server-side analogues of workflow data types"""

from abc import abstractmethod
from typing import Any, Dict, Generator, Generic, List, Type, TypeVar, Union

from sqlalchemy import select
from sqlalchemy.orm import Session

from .._db.datastore import workflow_db
from . import controller
from .asset import FIELDS, Asset

# Metadata
MetaType = TypeVar("MetaType", bound=controller.Record)

# Asset links
AssetLinkType = TypeVar("AssetLinkType", bound=controller.Record)


class DispatchedObject(Generic[MetaType, AssetLinkType]):
    """Base class for types with both metadata and assets.

    Each subclass must define two properties:
    - `meta_type`: The controller class for the type's "pure metadata" table
    - `asset_link_type`: The controller class for the type's asset-links table

    """

    @classmethod
    @property
    def meta_type(cls) -> Type[MetaType]:
        """Returns the metadata controller class."""
        raise NotImplementedError

    @classmethod
    @property
    def asset_link_type(cls) -> Type[AssetLinkType]:
        """Returns the asset link controller class"""
        raise NotImplementedError

    @classmethod
    @property
    def metadata_keys(cls) -> set:
        raise NotImplementedError

    @property
    @abstractmethod
    def query_keys(self) -> set:
        raise NotImplementedError

    @property
    @abstractmethod
    def metadata(self) -> MetaType:
        raise NotImplementedError

    @property
    def computed_fields(self) -> Dict:
        return {}

    @classmethod
    def session(cls) -> Generator[Session, None, None]:
        return workflow_db.session()

    def get_asset_ids(self, session: Session, keys: List[str]) -> Dict[str, int]:
        membership_filters = {"key": keys} if len(keys) > 0 else {}
        records = type(self).asset_link_type.get(
            session,
            fields=[],
            equality_filters={"meta_id": self._id},
            membership_filters=membership_filters,
        )
        return {x.key: x.asset_id for x in records}

    def associate_asset(
        self, session: Session, key: str, asset_id: int, flush: bool = False
    ) -> AssetLinkType:
        asset_link_kwargs = {
            "meta_id": self._id,
            "asset_id": asset_id,
            "key": key,
        }
        return type(self).asset_link_type.create(
            session, insert_kwargs=asset_link_kwargs, flush=flush
        )

    @property
    @abstractmethod
    def assets(self) -> Dict[str, Asset]:
        raise NotImplementedError

    @classmethod
    def meta_record_map(cls, key: str) -> str:
        return key

    def _refresh_metadata(self, session: Session, *, for_update: bool = False):
        fields = {type(self).meta_record_map(k) for k in self.query_keys}
        self.metadata.refresh(session, fields=fields, for_update=for_update)

    def get_metadata(self, key: str, session: Session, refresh: bool = True):
        attr = type(self).meta_record_map(key)
        if refresh:
            self._refresh_metadata(session)
        return self.metadata.attrs[attr]

    def set_metadata(self, key: str, val: Union[str, int], session: Session):
        record_attr = type(self).meta_record_map(key)
        self.metadata.update(session, values={record_attr: val})

    def incr_metadata(self, key: str, delta: int, session: Session):
        attr = type(self).meta_record_map(key)
        self.metadata.incr(session, increments={attr: delta})

    def get_asset(self, key: str, session: Session, refresh: bool = True) -> Asset:
        if key in self.assets:
            if refresh:
                self.assets[key].refresh(session, fields=FIELDS)
        else:
            asset_id = self.get_asset_ids(session, [key])[key]
            self.assets[key] = Asset.from_id(asset_id, session)

        return self.assets[key]

    def populate_asset_map(self, session: Session):
        """Load and cache all asset records"""
        asset_links = self.get_asset_ids(session=session, keys=[])
        for key, asset_id in asset_links.items():
            self.assets[key] = Asset.from_id(asset_id, session)

    def update_assets(self, updates: Dict[str, Dict], session: Session):
        """Bulk update associated assets"""
        for key, values in updates.items():
            asset = self.get_asset(key, session)
            asset.update(session, values=values)

    def _get_value(self, key: str, session: Session, refresh: bool = True) -> Any:
        if key in self.computed_fields:
            handler = self.computed_fields[key]
            return handler(self, session)
        elif key in self.metadata_keys:
            return self.get_metadata(key, session, refresh)
        else:
            return self.get_asset(key, session).load_data()

    def get_value(self, key: str, session: Session = None, refresh: bool = True) -> Any:
        if session is not None:
            return self._get_value(key, session, refresh)
        else:
            with self.session() as session:
                return self._get_value(key, session, refresh)

    def _set_value(self, key: str, val: Any, session: Session) -> None:
        if key in type(self).metadata_keys:
            self.set_metadata(key, val, session)
        else:
            self.get_asset(key, session).store_data(val, session)

    def set_value(self, key: str, val: Any, session: Session = None) -> None:
        if session is not None:
            self._set_value(key, val, session)
        else:
            with self.session() as session:
                self._set_value(key, val, session)

    def get_values(self, keys: List[str], session: Session = None, refresh: bool = True) -> Dict:
        return {key: self.get_value(key, session, refresh) for key in keys}

    @classmethod
    def get_db_records(
        cls,
        session: Session,
        *,
        keys: list,
        equality_filters: dict,
        membership_filters: dict,
        for_update: bool = False,
    ):
        # transform keys to db field names
        fields = list(map(cls.meta_record_map, keys))

        eq_filters_transformed = {}
        member_filters_transformed = {}
        for key, val in equality_filters.items():
            attr = cls.meta_record_map(key)
            eq_filters_transformed[attr] = val
        for key, vals in membership_filters.items():
            attr = cls.meta_record_map(key)
            member_filters_transformed[attr] = vals

        return cls.meta_type.get(
            session,
            fields=fields,
            equality_filters=eq_filters_transformed,
            membership_filters=member_filters_transformed,
            for_update=for_update,
        )

    @classmethod
    def get_linked_assets(
        cls, session, *, fields: list, equality_filters: dict, membership_filters: dict
    ) -> List[Asset]:
        link_model = cls.asset_link_type.model
        stmt = (
            select(link_model.meta_id, link_model.key, Asset.model)
            .join(link_model)
            .join(cls.meta_type.model)
        )
        fields = FIELDS
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.meta_type.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.meta_type.model, attr).in_(vals))

        attrs = [getattr(Asset.model, f) for f in fields]

        records = session.execute(stmt)

        return [
            {"meta_id": row.meta_id, "key": row.key, "asset": Asset(session, row[2], keys=fields)}
            for row in records
        ]



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import pickle
import uuid
from pathlib import Path

from fastapi import HTTPException
from fastapi.responses import JSONResponse
from sqlalchemy import extract, select
from sqlalchemy.sql import func

from covalent._results_manager.results_manager import get_result
from covalent._shared_files import logger
from covalent._shared_files.config import get_config
from covalent_dispatcher._core.execution import _get_task_inputs as get_task_inputs
from covalent_ui.api.v1.database.schema.electron import Electron
from covalent_ui.api.v1.database.schema.lattices import Lattice
from covalent_ui.api.v1.utils.file_handle import validate_data

app_log = logger.app_log
RESULTS_DIR = Path(get_config("dispatcher")["results_dir"]).resolve()


class Electrons:
    """Electron data access layer"""

    def __init__(self, db_con) -> None:
        self.db_con = db_con

    def electron_exist(self, electron_id: int) -> bool:
        return self.db_con.execute(
            select(Electron).where(Electron.transport_graph_node_id == electron_id)
        ).fetchone()

    def get_electrons_id(self, dispatch_id, electron_id) -> Electron:
        """
        Read electron table by electron id
        Args:
            electron_id: Refers to the electron's PK
        Return:
            Electron with PK as electron_id
        """
        data = (
            self.db_con.query(
                Electron.id,
                Electron.transport_graph_node_id,
                Electron.parent_lattice_id,
                Electron.type,
                Electron.storage_path,
                Electron.function_filename,
                Electron.function_string_filename,
                Electron.executor,
                Electron.executor_data,
                Electron.results_filename,
                Electron.value_filename,
                Electron.stdout_filename,
                Electron.hooks_filename,
                Electron.stderr_filename,
                Electron.error_filename,
                Electron.name,
                Electron.status,
                Electron.job_id,
                Electron.started_at.label("started_at"),
                Electron.completed_at.label("completed_at"),
                (
                    (
                        func.coalesce(
                            extract("epoch", Electron.completed_at),
                            extract("epoch", func.now()),
                        )
                        - extract("epoch", Electron.started_at)
                    )
                    * 1000
                ).label("runtime"),
            )
            .join(Lattice, Lattice.id == Electron.parent_lattice_id)
            .filter(
                Lattice.dispatch_id == str(dispatch_id),
                Electron.transport_graph_node_id == electron_id,
            )
            .first()
        )
        return data

    def get_electron_inputs(self, dispatch_id: uuid.UUID, electron_id: int) -> str:
        """
        Get Electron Inputs
        Args:
            dispatch_id: Dispatch id of lattice/sublattice
            electron_id: Transport graph node id of a electron
        Returns:
            Returns the inputs data from Result object
        """

        result = get_result(dispatch_id=str(dispatch_id), wait=False)
        if isinstance(result, JSONResponse) and result.status_code == 404:
            raise HTTPException(status_code=400, detail=result)
        result_object = pickle.loads(codecs.decode(result["result"].encode(), "base64"))
        electron_result = self.get_electrons_id(dispatch_id, electron_id)
        inputs = get_task_inputs(
            node_id=electron_id, node_name=electron_result.name, result_object=result_object
        )
        return validate_data(inputs)



# Copyright 2021 Agnostiq Inc.
#
# This file is part of Covalent.
#
# Licensed under the Apache License 2.0 (the "License"). A copy of the
# License may be obtained with this software package or at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Use of this file is prohibited except in compliance with the License.
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import annotations

from typing import Generic, List, Optional, Sequence, Type, TypeVar, Union

from sqlalchemy import delete, select, update
from sqlalchemy.engine import Row
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import Select, desc

from .._db import models

T = TypeVar("T", bound=models.Base)


class Record(Generic[T]):
    """
    Thin wrapper for a SQLALchemy record
    """

    @classmethod
    @property
    def model(cls) -> Type[T]:
        raise NotImplementedError

    def __init__(self, session: Session, record: models.Base, *, fields: list):
        self._id = record.id
        self._attrs = {k: getattr(record, k) for k in fields}

    @property
    def primary_key(self):
        return self._id

    @classmethod
    def get(
        cls,
        session: Session,
        *,
        stmt: Optional[Select] = None,
        fields: list,
        equality_filters: dict,
        membership_filters: dict,
        for_update: bool = False,
        sort_fields: List[str] = [],
        reverse: bool = True,
        offset: int = 0,
        max_items: Optional[int] = None,
    ) -> Union[Sequence[Row], Sequence[T]]:
        """Bulk ORM-enabled SELECT.

        Args:
            session: SQLalchemy session
            fields: List of columns to select
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
            for_update: Whether to lock the selected rows

        Returns:
            A list of SQLAlchemy Rows or whole ORM entities depending
        on whether only a subset of fields is specified.

        """
        if stmt is None:
            if len(fields) > 0:
                entities = [getattr(cls.model, attr) for attr in fields]
                stmt = select(*entities)
            else:
                stmt = select(cls.model)

        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        if for_update:
            stmt = stmt.with_for_update()
        for attr in sort_fields:
            if reverse:
                stmt = stmt.order_by(desc(getattr(cls.model, attr)))
            else:
                stmt = stmt.order_by(getattr(cls.model, attr))

        stmt = stmt.offset(offset)
        if max_items:
            stmt = stmt.limit(max_items)

        if len(fields) == 0:
            # Return whole ORM entities
            return session.scalars(stmt).all()
        else:
            # Return a named tuple containing the selected cols
            return session.execute(stmt).all()

    @classmethod
    def get_by_primary_key(
        cls, session: Session, primary_key: int, *, for_update: bool = False
    ) -> T:
        return session.get(cls.model, primary_key, with_for_update=for_update)

    @classmethod
    def create(cls, session: Session, *, insert_kwargs: dict, flush: bool = True) -> T:
        """Create a new record.

        Args:
            session: SQLalchemy session
            insert_kwargs: kwargs to pass to the model constructor
            flush: Whether to flush the session immediately

        Returns: A SQLAlchemy model of type T. If `flush=False`, the
            model will need to be added to the session manually.

        """

        new_record = cls.model(**insert_kwargs)
        session.add(new_record)
        if flush:
            session.flush()
        return new_record

    @classmethod
    def update_bulk(
        cls, session: Session, *, values: dict, equality_filters: dict, membership_filters: dict
    ):
        """Bulk update.

        Args:
            session: SQLAlchemy session
            values: dictionary of values to pass to UPDATE
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
        """

        stmt = update(cls.model).values(**values)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        session.execute(stmt)

    @classmethod
    def incr_bulk(
        cls,
        session: Session,
        *,
        increments: dict,
        equality_filters: dict,
        membership_filters: dict,
    ):
        """Bulk increment numerical fields

        Args:
            session: SQLAlchemy session
            increment: dictionary {field: delta}
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
        """

        kwargs = {}
        for field, delta in increments.items():
            col = getattr(cls.model, field)
            kwargs[field] = col + delta

        stmt = update(cls.model).values(**kwargs)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        session.execute(stmt)

    @classmethod
    def delete_bulk(
        cls,
        session: Session,
        *,
        equality_filters: dict,
        membership_filters: dict,
    ):
        """Bulk delete.

        Args:
            session: SQLAlchemy session
            equality_filters: Dict{field_name: value}
            membership_filters: Dict{field_name: value_list}
        """

        stmt = delete(cls.model)
        for attr, val in equality_filters.items():
            stmt = stmt.where(getattr(cls.model, attr) == val)
        for attr, vals in membership_filters.items():
            stmt = stmt.where(getattr(cls.model, attr).in_(vals))
        session.execute(stmt)

    def update(self, session: Session, *, values: dict):
        """Update the corresponding DB record."""

        type(self).update_bulk(
            session,
            values=values,
            equality_filters={"id": self.primary_key},
            membership_filters={},
        )
        for k, v in values.items():
            self._attrs[k] = v

    def incr(self, session: Session, *, increments: dict):
        """Increment the fields of the corresponding record."""
        type(self).incr_bulk(
            session,
            increments=increments,
            equality_filters={"id": self.primary_key},
            membership_filters={},
        )

    def refresh(self, session: Session, *, fields: set, for_update: bool = False):
        """Sync with DB"""
        records = type(self).get(
            session,
            fields=fields,
            equality_filters={"id": self._id},
            membership_filters={},
            for_update=for_update,
        )
        record = records[0]
        self._attrs = {k: getattr(record, k) for k in fields}

    @property
    def attrs(self) -> dict:
        return self._attrs

    def __contains__(self, item: str):
        return item in self._attrs
